@@ -1,4 +1,5 @@
 #include "arch.h"
+#include "utils.h"
 
 #include <algorithm>
 #include <array>
@@ -49,8 +50,7 @@ static std::array<std::string, ARG_REGISTERS> arg_registers = {
 
 int offset(std::string reg_name)
 {
-  for (unsigned int i = 0; i < registers.size(); i++)
-  {
+  for (unsigned int i = 0; i < registers.size(); i++) {
     if (registers[i].count(reg_name))
       return i;
   }
@@ -94,7 +94,7 @@ std::string name()
 
 std::vector<std::string> invalid_watchpoint_modes()
 {
-  throw std::runtime_error(
+  throw FatalUserException(
       "Watchpoints are not supported on this architecture");
 }
 
