@@ -6,7 +6,6 @@
 #include <vector>
 
 #include "location.hh"
-#include "mapkey.h"
 #include "types.h"
 #include "usdt.h"
 #include "utils.h"
@@ -14,19 +13,7 @@
 namespace bpftrace {
 namespace ast {
 
-class VisitorBase;
-
-#define DEFINE_ACCEPT void accept(VisitorBase &v) override;
-
-/**
- * Copy the node but leave all it's child members uninitialized, effectively
- * turning it into a leaf node
- */
-#define DEFINE_LEAFCOPY(T)                                                     \
-  T *leafcopy()                                                                \
-  {                                                                            \
-    return new T(*this);                                                       \
-  };
+class ASTContext;
 
 enum class JumpType {
   INVALID = 0,
@@ -75,18 +62,14 @@ enum class ExpansionType {
 
 class Node {
 public:
-  Node() = default;
-  Node(location loc) : loc(loc){};
-  Node(const Node &other) = default;
+  Node(location loc) : loc(loc) {};
+  virtual ~Node() = default;
 
+  Node(const Node &) = delete;
   Node &operator=(const Node &) = delete;
   Node(Node &&) = delete;
   Node &operator=(Node &&) = delete;
 
-  virtual ~Node() = default;
-
-  virtual void accept(VisitorBase &v) = 0;
-
   location loc;
 };
 
@@ -94,16 +77,7 @@ class Map;
 class Variable;
 class Expression : public Node {
 public:
-  Expression() = default;
-  Expression(location loc) : Node(loc){};
-  Expression(const Expression &other);
-
-  Expression &operator=(const Expression &) = delete;
-  Expression(Expression &&) = delete;
-  Expression &operator=(Expression &&) = delete;
-
-  // NB: do not free any of non-owned pointers we store
-  virtual ~Expression() = default;
+  Expression(location loc) : Node(loc) {};
 
   SizedType type;
   Map *key_for_map = nullptr;
@@ -117,84 +91,47 @@ using ExpressionList = std::vector<Expression *>;
 
 class Integer : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Integer)
-
-  explicit Integer(int64_t n, location loc);
+  explicit Integer(int64_t n, location loc, bool is_negative = true);
 
   int64_t n;
-
-private:
-  Integer(const Integer &other) = default;
+  bool is_negative;
 };
 
 class PositionalParameter : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(PositionalParameter)
-
   explicit PositionalParameter(PositionalParameterType ptype,
                                long n,
                                location loc);
-  ~PositionalParameter() = default;
 
   PositionalParameterType ptype;
   long n;
   bool is_in_str = false;
-
-private:
-  PositionalParameter(const PositionalParameter &other) = default;
 };
 
 class String : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(String)
-
   explicit String(const std::string &str, location loc);
-  ~String() = default;
 
   std::string str;
-
-private:
-  String(const String &other) = default;
 };
 
 class StackMode : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(StackMode)
-
   explicit StackMode(const std::string &mode, location loc);
-  ~StackMode() = default;
 
   std::string mode;
-
-private:
-  StackMode(const StackMode &other) = default;
 };
 
 class Identifier : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Identifier)
-
   explicit Identifier(const std::string &ident, location loc);
-  ~Identifier() = default;
 
   std::string ident;
-
-private:
-  Identifier(const Identifier &other) = default;
 };
 
 class Builtin : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Builtin)
-
   explicit Builtin(const std::string &ident, location loc);
-  ~Builtin() = default;
 
   std::string ident;
   int probe_id;
@@ -205,351 +142,213 @@ public:
     return !ident.compare(0, 3, "arg") && ident.size() == 4 &&
            ident.at(3) >= '0' && ident.at(3) <= '9';
   }
-
-private:
-  Builtin(const Builtin &other) = default;
 };
 
 class Call : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Call)
-
   explicit Call(const std::string &func, location loc);
-  Call(const std::string &func, ExpressionList *vargs, location loc);
-  ~Call();
+  Call(const std::string &func, ExpressionList &&vargs, location loc);
 
   std::string func;
-  ExpressionList *vargs = nullptr;
-
-private:
-  Call(const Call &other);
+  ExpressionList vargs;
 };
 
 class Sizeof : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Sizeof)
-
   Sizeof(SizedType type, location loc);
   Sizeof(Expression *expr, location loc);
-  ~Sizeof();
 
-  Expression *expr;
+  Expression *expr = nullptr;
   SizedType argtype;
-
-private:
-  Sizeof(const Sizeof &other);
 };
 
 class Offsetof : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Offsetof)
-
-  Offsetof(SizedType record, std::string &field, location loc);
-  Offsetof(Expression *expr, std::string &field, location loc);
-  ~Offsetof();
+  Offsetof(SizedType record, std::vector<std::string> &field, location loc);
+  Offsetof(Expression *expr, std::vector<std::string> &field, location loc);
 
   SizedType record;
-  Expression *expr;
-  std::string field;
-
-private:
-  Offsetof(const Offsetof &other);
+  Expression *expr = nullptr;
+  std::vector<std::string> field;
 };
 
 class Map : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Map)
-
   explicit Map(const std::string &ident, location loc);
-  Map(const std::string &ident, ExpressionList *vargs, location loc);
-  ~Map();
+  Map(const std::string &ident, Expression &expr, location loc);
 
   std::string ident;
-  MapKey key_type;
-  ExpressionList *vargs = nullptr;
+  Expression *key_expr = nullptr;
+  SizedType key_type;
   bool skip_key_validation = false;
-
-private:
-  Map(const Map &other);
+  // This is for a feature check on reading per-cpu maps
+  // which involve calling map_lookup_percpu_elem
+  // https://github.com/bpftrace/bpftrace/issues/3755
+  bool is_read = true;
 };
 
 class Variable : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Variable)
-
   explicit Variable(const std::string &ident, location loc);
-  ~Variable() = default;
 
   std::string ident;
-
-private:
-  Variable(const Variable &other) = default;
 };
 
 class Binop : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Binop)
-
   Binop(Expression *left, Operator op, Expression *right, location loc);
 
-  ~Binop();
-
   Expression *left = nullptr;
   Expression *right = nullptr;
   Operator op;
-
-private:
-  Binop(const Binop &other);
 };
 
 class Unop : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Unop)
-
-  Unop(Operator op, Expression *expr, location loc = location());
-  Unop(Operator op,
-       Expression *expr,
-       bool is_post_op = false,
-       location loc = location());
-
-  ~Unop();
+  Unop(Operator op, Expression *expr, bool is_post_op, location loc);
 
   Expression *expr = nullptr;
   Operator op;
   bool is_post_op;
-
-private:
-  Unop(const Unop &other);
 };
 
 class FieldAccess : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(FieldAccess)
-
   FieldAccess(Expression *expr, const std::string &field);
   FieldAccess(Expression *expr, const std::string &field, location loc);
   FieldAccess(Expression *expr, ssize_t index, location loc);
-  ~FieldAccess();
 
   Expression *expr = nullptr;
   std::string field;
   ssize_t index = -1;
-
-private:
-  FieldAccess(const FieldAccess &other);
 };
 
 class ArrayAccess : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(ArrayAccess)
-
   ArrayAccess(Expression *expr, Expression *indexpr);
   ArrayAccess(Expression *expr, Expression *indexpr, location loc);
-  ~ArrayAccess();
 
   Expression *expr = nullptr;
   Expression *indexpr = nullptr;
-
-private:
-  ArrayAccess(const ArrayAccess &other) : Expression(other){};
 };
 
 class Cast : public Expression {
 public:
-  DEFINE_LEAFCOPY(Cast)
-  DEFINE_ACCEPT
-
   Cast(SizedType type, Expression *expr, location loc);
-  ~Cast();
 
   Expression *expr = nullptr;
-
-private:
-  Cast(const Cast &other);
 };
 
 class Tuple : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Tuple)
+  Tuple(ExpressionList &&elems, location loc);
 
-  Tuple(ExpressionList *elems, location loc);
-  ~Tuple();
-
-  ExpressionList *elems = nullptr;
-
-private:
-  Tuple(const Tuple &other);
+  ExpressionList elems;
 };
 
 class Statement : public Node {
 public:
-  Statement() = default;
-  Statement(location loc) : Node(loc){};
-  Statement(const Statement &other) = default;
-  ~Statement() = default;
-
-  Statement &operator=(const Statement &) = delete;
-  Statement(Statement &&) = delete;
-  Statement &operator=(Statement &&) = delete;
+  Statement(location loc) : Node(loc) {};
 };
 
 using StatementList = std::vector<Statement *>;
 
 class ExprStatement : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(ExprStatement)
-
   explicit ExprStatement(Expression *expr, location loc);
-  ~ExprStatement();
 
   Expression *expr = nullptr;
+};
 
-private:
-  ExprStatement(const ExprStatement &other) : Statement(other){};
+class VarDeclStatement : public Statement {
+public:
+  VarDeclStatement(Variable *var, SizedType type, location loc);
+  VarDeclStatement(Variable *var, location loc);
+
+  Variable *var = nullptr;
+  bool set_type = false;
 };
 
 class AssignMapStatement : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(AssignMapStatement)
-
-  AssignMapStatement(Map *map,
-                     Expression *expr,
-                     bool compound = false,
-                     location loc = location());
-  ~AssignMapStatement();
+  AssignMapStatement(Map *map, Expression *expr, location loc);
 
   Map *map = nullptr;
   Expression *expr = nullptr;
-  bool compound;
-
-private:
-  AssignMapStatement(const AssignMapStatement &other);
 };
 
 class AssignVarStatement : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(AssignVarStatement)
-
-  AssignVarStatement(Variable *var,
+  AssignVarStatement(Variable *var, Expression *expr, location loc);
+  AssignVarStatement(VarDeclStatement *var_decl_stmt,
                      Expression *expr,
-                     bool compound = false,
-                     location loc = location());
-  ~AssignVarStatement();
+                     location loc);
 
+  VarDeclStatement *var_decl_stmt = nullptr;
   Variable *var = nullptr;
   Expression *expr = nullptr;
-  bool compound;
-
-private:
-  AssignVarStatement(const AssignVarStatement &other);
 };
 
 class AssignConfigVarStatement : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(AssignConfigVarStatement)
-
   AssignConfigVarStatement(const std::string &config_var,
                            Expression *expr,
-                           location loc = location());
-  ~AssignConfigVarStatement();
+                           location loc);
 
   std::string config_var;
   Expression *expr = nullptr;
+};
 
-private:
-  AssignConfigVarStatement(const AssignConfigVarStatement &other);
+class Block : public Statement {
+public:
+  Block(StatementList &&stmts, location loc);
+
+  StatementList stmts;
 };
 
 class If : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(If)
-
-  If(Expression *cond, StatementList *stmts);
-  If(Expression *cond, StatementList *stmts, StatementList *else_stmts);
-  ~If();
+  If(Expression *cond, Block *if_block, Block *else_block, location loc);
 
   Expression *cond = nullptr;
-  StatementList *stmts = nullptr;
-  StatementList *else_stmts = nullptr;
-
-private:
-  If(const If &other);
+  Block *if_block = nullptr;
+  Block *else_block = nullptr;
 };
 
 class Unroll : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Unroll)
-
-  Unroll(Expression *expr, StatementList *stmts, location loc);
-  ~Unroll();
+  Unroll(Expression *expr, Block *block, location loc);
 
   long int var = 0;
   Expression *expr = nullptr;
-  StatementList *stmts = nullptr;
-
-private:
-  Unroll(const Unroll &other);
+  Block *block = nullptr;
 };
 
 class Jump : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Jump)
-
-  Jump(JumpType ident, Expression *return_value, location loc = location())
+  Jump(JumpType ident, Expression *return_value, location loc)
       : Statement(loc), ident(ident), return_value(return_value)
   {
   }
-  Jump(JumpType ident, location loc = location())
+  Jump(JumpType ident, location loc)
       : Statement(loc), ident(ident), return_value(nullptr)
   {
   }
-  ~Jump();
 
   JumpType ident = JumpType::INVALID;
   Expression *return_value;
-
-private:
-  Jump(const Jump &other) = default;
 };
 
 class Predicate : public Node {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Predicate)
-
   explicit Predicate(Expression *expr, location loc);
-  ~Predicate();
 
   Expression *expr = nullptr;
-
-private:
-  Predicate(const Predicate &other) : Node(other){};
 };
 
 class Ternary : public Expression {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Ternary)
-
   Ternary(Expression *cond, Expression *left, Expression *right, location loc);
-  ~Ternary();
 
   Expression *cond = nullptr;
   Expression *left = nullptr;
@@ -558,77 +357,46 @@ public:
 
 class While : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(While)
-
-  While(Expression *cond, StatementList *stmts, location loc)
-      : Statement(loc), cond(cond), stmts(stmts)
+  While(Expression *cond, Block *block, location loc)
+      : Statement(loc), cond(cond), block(block)
   {
   }
-  ~While();
 
   Expression *cond = nullptr;
-  StatementList *stmts = nullptr;
-
-private:
-  While(const While &other);
+  Block *block = nullptr;
 };
 
 class For : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(For)
-
-  For(Variable *decl, Expression *expr, StatementList *stmts, location loc)
-      : Statement(loc), decl(decl), expr(expr), stmts(stmts)
+  For(Variable *decl, Expression *expr, StatementList &&stmts, location loc)
+      : Statement(loc), decl(decl), expr(expr), stmts(std::move(stmts))
   {
   }
-  ~For();
 
   Variable *decl = nullptr;
   Expression *expr = nullptr;
-  StatementList *stmts = nullptr;
-
-private:
-  For(const For &other);
+  StatementList stmts;
+  SizedType ctx_type;
 };
 
 class Config : public Statement {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Config)
+  Config(StatementList &&stmts, location loc)
+      : Statement(loc), stmts(std::move(stmts)) {};
 
-  Config(StatementList *stmts) : stmts(stmts)
-  {
-  }
-  ~Config();
-
-  StatementList *stmts = nullptr;
-
-private:
-  Config(const Config &other);
-};
-
-class Scope : public Node {
-public:
-  Scope(StatementList *stmts);
-  virtual ~Scope() = 0;
-
-  StatementList *stmts;
+  StatementList stmts;
 };
 
 class AttachPoint : public Node {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(AttachPoint)
+  AttachPoint(const std::string &raw_input, bool ignore_invalid, location loc);
 
-  explicit AttachPoint(const std::string &raw_input, location loc = location());
-  AttachPoint(const std::string &raw_input, bool ignore_invalid)
-      : AttachPoint(raw_input)
-  {
-    this->ignore_invalid = ignore_invalid;
-  }
-  ~AttachPoint() = default;
+  // Currently, the AST node itself is used to store metadata related to probe
+  // expansion and attachment. This is done through `create_expansion_copy`
+  // below.  Since the nodes are not currently copyable by default (this is
+  // currently fraught, as nodes may have backreferences that are not updated
+  // in these cases), these fields are copied manually. *Until this is fixed,
+  // if you are adding new fields, be sure to update `create_expansion_copy`.
 
   // Raw, unparsed input from user, eg. kprobe:vfs_read
   std::string raw_input;
@@ -654,28 +422,27 @@ public:
 
   std::string name() const;
 
-  AttachPoint create_expansion_copy(const std::string &match) const;
+  AttachPoint &create_expansion_copy(ASTContext &ctx,
+                                     const std::string &match) const;
 
   int index() const;
   void set_index(int index);
 
 private:
-  AttachPoint(const AttachPoint &other) = default;
-
   int index_ = 0;
 };
 using AttachPointList = std::vector<AttachPoint *>;
 
-class Probe : public Scope {
+class Probe : public Node {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Probe)
+  Probe(AttachPointList &&attach_points,
+        Predicate *pred,
+        Block *block,
+        location loc);
 
-  Probe(AttachPointList *attach_points, Predicate *pred, StatementList *stmts);
-  ~Probe();
-
-  AttachPointList *attach_points = nullptr;
+  AttachPointList attach_points;
   Predicate *pred = nullptr;
+  Block *block = nullptr;
 
   std::string name() const;
   std::string args_typename() const;
@@ -689,67 +456,53 @@ public:
   bool has_ap_of_probetype(ProbeType probe_type);
 
 private:
-  Probe(const Probe &other);
   int index_ = 0;
 };
 using ProbeList = std::vector<Probe *>;
 
 class SubprogArg : public Node {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(SubprogArg)
-
-  SubprogArg(std::string name, SizedType type);
+  SubprogArg(std::string name, SizedType type, location loc);
 
   std::string name() const;
   SizedType type;
 
 private:
-  SubprogArg(const SubprogArg &other);
   std::string name_;
 };
 using SubprogArgList = std::vector<SubprogArg *>;
 
-class Subprog : public Scope {
+class Subprog : public Node {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Subprog)
-
   Subprog(std::string name,
           SizedType return_type,
-          SubprogArgList *args,
-          StatementList *stmts);
-  ~Subprog();
+          SubprogArgList &&args,
+          StatementList &&stmts,
+          location loc);
 
-  SubprogArgList *args = nullptr;
+  SubprogArgList args;
   SizedType return_type;
+  StatementList stmts;
 
   std::string name() const;
 
 private:
-  Subprog(const Subprog &other);
   std::string name_;
 };
 using SubprogList = std::vector<Subprog *>;
 
 class Program : public Node {
 public:
-  DEFINE_ACCEPT
-  DEFINE_LEAFCOPY(Program)
-
   Program(const std::string &c_definitions,
           Config *config,
-          SubprogList *functions,
-          ProbeList *probes);
-  ~Program();
+          SubprogList &&functions,
+          ProbeList &&probes,
+          location loc);
 
   std::string c_definitions;
   Config *config = nullptr;
-  SubprogList *functions = nullptr;
-  ProbeList *probes = nullptr;
-
-private:
-  Program(const Program &other);
+  SubprogList functions;
+  ProbeList probes;
 };
 
 std::string opstr(const Binop &binop);
@@ -757,8 +510,37 @@ std::string opstr(const Unop &unop);
 std::string opstr(const Jump &jump);
 
 SizedType ident_to_record(const std::string &ident, int pointer_level = 0);
+SizedType ident_to_sized_type(const std::string &ident);
 
-#undef DEFINE_ACCEPT
+template <typename T>
+concept NodeType = std::derived_from<T, Node>;
+
+// Manages the lifetime of AST nodes.
+//
+// Nodes allocated by an ASTContext will be kept alive for the duration of the
+// owning ASTContext object.
+class ASTContext {
+public:
+  Program *root = nullptr;
+
+  // Creates and returns a pointer to an AST node.
+  template <NodeType T, typename... Args>
+  T *make_node(Args &&...args)
+  {
+    auto uniq_ptr = std::make_unique<T>(std::forward<Args>(args)...);
+    auto *raw_ptr = uniq_ptr.get();
+    nodes_.push_back(std::move(uniq_ptr));
+    return raw_ptr;
+  }
+
+  unsigned int node_count()
+  {
+    return nodes_.size();
+  }
+
+private:
+  std::vector<std::unique_ptr<Node>> nodes_;
+};
 
 } // namespace ast
 } // namespace bpftrace
