@@ -13,8 +13,7 @@
 #include <unordered_map>
 #include <vector>
 
-namespace bpftrace {
-namespace ast {
+namespace bpftrace::ast {
 
 AttachPointParser::State AttachPointParser::argument_count_error(
     int expected,
@@ -52,23 +51,23 @@ std::optional<int64_t> AttachPointParser::stoll(const std::string &str)
   }
 }
 
-AttachPointParser::AttachPointParser(Program *root,
+AttachPointParser::AttachPointParser(ASTContext &ctx,
                                      BPFtrace &bpftrace,
                                      std::ostream &sink,
                                      bool listing)
-    : root_(root), bpftrace_(bpftrace), sink_(sink), listing_(listing)
+    : ctx_(ctx), bpftrace_(bpftrace), sink_(sink), listing_(listing)
 {
 }
 
 int AttachPointParser::parse()
 {
-  if (!root_)
+  if (!ctx_.root)
     return 1;
 
   uint32_t failed = 0;
-  for (Probe *probe : *(root_->probes)) {
-    for (size_t i = 0; i < probe->attach_points->size(); ++i) {
-      auto ap_ptr = (*probe->attach_points)[i];
+  for (Probe *probe : ctx_.root->probes) {
+    for (size_t i = 0; i < probe->attach_points.size(); ++i) {
+      auto ap_ptr = probe->attach_points[i];
       auto &ap = *ap_ptr;
       new_attach_points.clear();
 
@@ -78,14 +77,13 @@ int AttachPointParser::parse()
         LOG(ERROR, ap.loc, sink_) << errs_.str();
       } else if (s == SKIP || s == NEW_APS) {
         // Remove the current attach point
-        delete ap_ptr;
-        probe->attach_points->erase(probe->attach_points->begin() + i);
+        probe->attach_points.erase(probe->attach_points.begin() + i);
         i--;
         if (s == NEW_APS) {
           // The removed attach point is replaced by new ones
-          probe->attach_points->insert(probe->attach_points->end(),
-                                       new_attach_points.begin(),
-                                       new_attach_points.end());
+          probe->attach_points.insert(probe->attach_points.end(),
+                                      new_attach_points.begin(),
+                                      new_attach_points.end());
         }
       }
 
@@ -94,14 +92,14 @@ int AttachPointParser::parse()
       errs_.str({});
     }
 
-    auto new_end = std::remove_if(probe->attach_points->begin(),
-                                  probe->attach_points->end(),
+    auto new_end = std::remove_if(probe->attach_points.begin(),
+                                  probe->attach_points.end(),
                                   [](const AttachPoint *ap) {
                                     return ap->provider.empty();
                                   });
-    probe->attach_points->erase(new_end, probe->attach_points->end());
+    probe->attach_points.erase(new_end, probe->attach_points.end());
 
-    if (probe->attach_points->empty()) {
+    if (probe->attach_points.empty()) {
       LOG(ERROR, probe->loc, sink_) << "No attach points for probe";
       failed++;
     }
@@ -165,8 +163,9 @@ AttachPointParser::State AttachPointParser::parse_attachpoint(AttachPoint &ap)
       raw_input = probe_type + ":" + raw_input;
       // New attach points have ignore_invalid set to true - probe types for
       // which raw_input has invalid number of parts will be ignored (instead
-      // of throwing an error)
-      new_attach_points.push_back(new AttachPoint(raw_input, true));
+      // of throwing an error). These will have the same associated location.
+      new_attach_points.push_back(
+          ctx_.make_node<AttachPoint>(raw_input, true, ap.loc));
     }
     return NEW_APS;
   }
@@ -200,9 +199,9 @@ AttachPointParser::State AttachPointParser::parse_attachpoint(AttachPoint &ap)
       return watchpoint_parser();
     case ProbeType::asyncwatchpoint:
       return watchpoint_parser(true);
-    case ProbeType::kfunc:
-    case ProbeType::kretfunc:
-      return kfunc_parser();
+    case ProbeType::fentry:
+    case ProbeType::fexit:
+      return fentry_parser();
     case ProbeType::iter:
       return iter_parser();
     case ProbeType::rawtracepoint:
@@ -278,13 +277,22 @@ AttachPointParser::State AttachPointParser::lex_attachpoint(
 
 AttachPointParser::State AttachPointParser::special_parser()
 {
-  // Can only have reached here if provider is `BEGIN` or `END`
-  assert(ap_->provider == "BEGIN" || ap_->provider == "END");
+  // Can only have reached here if provider is `BEGIN` or `END` or `self`
+  assert(ap_->provider == "BEGIN" || ap_->provider == "END" ||
+         ap_->provider == "self");
 
-  if (parts_.size() == 2 && parts_[1] == "*")
-    parts_.pop_back();
-  if (parts_.size() != 1) {
-    return argument_count_error(0);
+  if (ap_->provider == "BEGIN" || ap_->provider == "END") {
+    if (parts_.size() == 2 && parts_[1] == "*")
+      parts_.pop_back();
+    if (parts_.size() != 1) {
+      return argument_count_error(0);
+    }
+  } else if (ap_->provider == "self") {
+    if (parts_.size() != 3) {
+      return argument_count_error(2);
+    }
+    ap_->target = parts_[1];
+    ap_->func = parts_[2];
   }
 
   return OK;
@@ -392,10 +400,11 @@ AttachPointParser::State AttachPointParser::uprobe_parser(bool allow_offset,
     // If the target has form "libXXX" then we use BCC to find the correct path
     // to the given library as it may differ across systems.
     auto libname = parts_[1].substr(3);
-    const char *lib_path = bcc_procutils_which_so(libname.c_str(),
-                                                  bpftrace_.pid());
-    if (lib_path)
+    auto lib_path = bcc_procutils_which_so(libname.c_str(), bpftrace_.pid());
+    if (lib_path) {
       ap_->target = lib_path;
+      ::free(lib_path);
+    }
   }
 
   if (ap_->target.empty()) {
@@ -676,9 +685,9 @@ AttachPointParser::State AttachPointParser::watchpoint_parser(bool async)
   return OK;
 }
 
-AttachPointParser::State AttachPointParser::kfunc_parser()
+AttachPointParser::State AttachPointParser::fentry_parser()
 {
-  // kfunc[:module]:function
+  // fentry[:module]:function
   if (parts_.size() != 2 && parts_.size() != 3) {
     if (ap_->ignore_invalid)
       return SKIP;
@@ -766,5 +775,4 @@ AttachPointParser::State AttachPointParser::raw_tracepoint_parser()
   return OK;
 }
 
-} // namespace ast
-} // namespace bpftrace
+} // namespace bpftrace::ast
