@@ -8,41 +8,12 @@
 #include <optional>
 
 #include "ast/ast.h"
+#include "ast/async_ids.h"
 #include "bpftrace.h"
 #include "types.h"
 
-#if LLVM_VERSION_MAJOR >= 5 && LLVM_VERSION_MAJOR < 7
-#define CREATE_MEMCPY(dst, src, size, algn)                                    \
-  CreateMemCpy((dst), (src), (size), (algn))
-#define CREATE_MEMCPY_VOLATILE(dst, src, size, algn)                           \
-  CreateMemCpy((dst), (src), (size), (algn), true)
-#elif LLVM_VERSION_MAJOR >= 7 && LLVM_VERSION_MAJOR < 10
-#define CREATE_MEMCPY(dst, src, size, algn)                                    \
-  CreateMemCpy((dst), (algn), (src), (algn), (size))
-#define CREATE_MEMCPY_VOLATILE(dst, src, size, algn)                           \
-  CreateMemCpy((dst), (algn), (src), (algn), (size), true)
-#elif LLVM_VERSION_MAJOR >= 10
-#define CREATE_MEMCPY(dst, src, size, algn)                                    \
-  CreateMemCpy((dst), MaybeAlign(algn), (src), MaybeAlign(algn), (size))
-#define CREATE_MEMCPY_VOLATILE(dst, src, size, algn)                           \
-  CreateMemCpy((dst), MaybeAlign(algn), (src), MaybeAlign(algn), (size), true)
-#else
-#error Unsupported LLVM version
-#endif
-
-#if LLVM_VERSION_MAJOR >= 13
 #define CREATE_ATOMIC_RMW(op, ptr, val, align, order)                          \
   CreateAtomicRMW((op), (ptr), (val), MaybeAlign((align)), (order))
-#else
-#define CREATE_ATOMIC_RMW(op, ptr, val, align, order)                          \
-  CreateAtomicRMW((op), (ptr), (val), (order))
-#endif
-
-#if LLVM_VERSION_MAJOR >= 15
-#define GET_PTR_TY() getPtrTy()
-#else
-#define GET_PTR_TY() getInt8PtrTy()
-#endif
 
 namespace bpftrace {
 namespace ast {
@@ -51,22 +22,20 @@ using namespace llvm;
 
 class IRBuilderBPF : public IRBuilder<> {
 public:
-  IRBuilderBPF(LLVMContext &context, Module &module, BPFtrace &bpftrace);
+  IRBuilderBPF(LLVMContext &context,
+               Module &module,
+               BPFtrace &bpftrace,
+               AsyncIds &async_ids);
 
   AllocaInst *CreateAllocaBPF(llvm::Type *ty, const std::string &name = "");
   AllocaInst *CreateAllocaBPF(const SizedType &stype,
                               const std::string &name = "");
   AllocaInst *CreateAllocaBPFInit(const SizedType &stype,
                                   const std::string &name);
-  AllocaInst *CreateAllocaBPF(llvm::Type *ty,
-                              llvm::Value *arraysize,
-                              const std::string &name = "");
-  AllocaInst *CreateAllocaBPF(const SizedType &stype,
-                              llvm::Value *arraysize,
-                              const std::string &name = "");
   AllocaInst *CreateAllocaBPF(int bytes, const std::string &name = "");
   void CreateMemsetBPF(Value *ptr, Value *val, uint32_t size);
-  llvm::Type *GetType(const SizedType &stype);
+  void CreateMemcpyBPF(Value *dst, Value *src, uint32_t size);
+  llvm::Type *GetType(const SizedType &stype, bool emit_codegen_types = true);
   llvm::Type *GetMapValueType(const SizedType &stype);
   llvm::ConstantInt *GetIntSameSize(uint64_t C, llvm::Value *expr);
   llvm::ConstantInt *GetIntSameSize(uint64_t C, llvm::Type *ty);
@@ -88,8 +57,7 @@ public:
                                  Map &map,
                                  Value *key,
                                  const SizedType &type,
-                                 const location &loc,
-                                 bool is_aot);
+                                 const location &loc);
   void CreateMapUpdateElem(Value *ctx,
                            const std::string &map_ident,
                            Value *key,
@@ -100,11 +68,11 @@ public:
                            Map &map,
                            Value *key,
                            const location &loc);
-  void CreateForEachMapElem(Value *ctx,
-                            Map &map,
-                            Value *callback,
-                            Value *callback_ctx,
-                            const location &loc);
+  Value *CreateForEachMapElem(Value *ctx,
+                              Map &map,
+                              Value *callback,
+                              Value *callback_ctx,
+                              const location &loc);
   void CreateProbeRead(Value *ctx,
                        Value *dst,
                        llvm::Value *size,
@@ -155,15 +123,7 @@ public:
   Value *CreateStrcontains(Value *val1,
                            uint64_t str1_size,
                            Value *val2,
-                           uint64_t str2_size,
-                           bool inverse);
-  Value *CreateIntegerArrayCmpUnrolled(Value *ctx,
-                                       Value *val1,
-                                       Value *val2,
-                                       const SizedType &val1_type,
-                                       const SizedType &val2_type,
-                                       const bool inverse,
-                                       const location &loc);
+                           uint64_t str2_size);
   Value *CreateIntegerArrayCmp(Value *ctx,
                                Value *val1,
                                Value *val2,
@@ -174,7 +134,6 @@ public:
                                MDNode *metadata);
   CallInst *CreateGetNs(TimestampMode ts, const location &loc);
   CallInst *CreateJiffies64(const location &loc);
-  CallInst *CreateGetPidTgid(const location &loc);
   CallInst *CreateGetCurrentCgroupId(const location &loc);
   CallInst *CreateGetUidGid(const location &loc);
   CallInst *CreateGetNumaId(const location &loc);
@@ -187,13 +146,28 @@ public:
                            StackType stack_type,
                            const location &loc);
   CallInst *CreateGetFuncIp(Value *ctx, const location &loc);
+  CallInst *CreatePerCpuPtr(Value *var, Value *cpu, const location &loc);
+  CallInst *CreateThisCpuPtr(Value *var, const location &loc);
   CallInst *CreateGetJoinMap(BasicBlock *failure_callback, const location &loc);
   CallInst *CreateGetStackScratchMap(StackType stack_type,
                                      BasicBlock *failure_callback,
                                      const location &loc);
-  CallInst *CreateGetStrScratchMap(int idx,
-                                   BasicBlock *failure_callback,
-                                   const location &loc);
+  Value *CreateGetStrAllocation(const std::string &name, const location &loc);
+  Value *CreateGetFmtStringArgsAllocation(StructType *struct_type,
+                                          const std::string &name,
+                                          const location &loc);
+  Value *CreateTupleAllocation(const SizedType &tuple_type,
+                               const std::string &name,
+                               const location &loc);
+  Value *CreateWriteMapValueAllocation(const SizedType &value_type,
+                                       const std::string &name,
+                                       const location &loc);
+  Value *CreateVariableAllocationInit(const SizedType &value_type,
+                                      const std::string &name,
+                                      const location &loc);
+  Value *CreateMapKeyAllocation(const SizedType &value_type,
+                                const std::string &name,
+                                const location &loc);
   void CreateCheckSetRecursion(const location &loc, int early_exit_ret);
   void CreateUnSetRecursion(const location &loc);
   CallInst *CreateHelperCall(libbpf::bpf_func_id func_id,
@@ -242,13 +216,21 @@ public:
                              libbpf::bpf_func_id func_id,
                              const location &loc,
                              bool compare_zero = false);
+  StructType *GetStackStructType(bool is_ustack);
   StructType *GetStructType(std::string name,
                             const std::vector<llvm::Type *> &elements,
                             bool packed = false);
-  AllocaInst *CreateUSym(llvm::Value *val, int probe_id, const location &loc);
+  Value *CreateGetPid(const location &loc);
+  Value *CreateGetTid(const location &loc);
+  Value *CreateGetPid(Value *ctx, const location &loc);
+  Value *CreateGetTid(Value *ctx, const location &loc);
+  AllocaInst *CreateUSym(Value *ctx,
+                         Value *val,
+                         int probe_id,
+                         const location &loc);
   Value *CreateRegisterRead(Value *ctx, const std::string &builtin);
   Value *CreateRegisterRead(Value *ctx, int offset, const std::string &name);
-  Value *CreatKFuncArg(Value *ctx, SizedType &type, std::string &name);
+  Value *CreateKFuncArg(Value *ctx, SizedType &type, std::string &name);
   Value *CreateRawTracepointArg(Value *ctx, const std::string &builtin);
   Value *CreateUprobeArgsRecord(Value *ctx, const SizedType &args_type);
   llvm::Type *UprobeArgsType(const SizedType &args_type);
@@ -256,7 +238,11 @@ public:
                             Value *len,
                             AllocaInst *data,
                             size_t size);
-  void CreatePath(Value *ctx, Value *buf, Value *path, const location &loc);
+  void CreatePath(Value *ctx,
+                  Value *buf,
+                  Value *path,
+                  Value *sz,
+                  const location &loc);
   void CreateSeqPrintf(Value *ctx,
                        Value *fmt,
                        Value *fmt_size,
@@ -271,6 +257,20 @@ public:
                                llvm::Value *index,
                                AddrSpace as);
 
+  // Safely handle pointer references by wrapping the address with the
+  // intrinsic `preserve_static_offset` [1], which will ensure that LLVM does
+  // not apply certain basic optimizations (notably, saving any intermediate
+  // offset from this pointer). This is required for the context pointer,
+  // which, if modified. will trigger an error in the verifier. This method
+  // also automatically handles casts from integers and other pointers; the
+  // output value is always a pointer to `ty`.
+  //
+  // [1] https://reviews.llvm.org/D133361
+  llvm::Value *CreateSafeGEP(llvm::Type *Ty,
+                             llvm::Value *Ptr,
+                             llvm::ArrayRef<Value *> IdxList,
+                             const llvm::Twine &Name = "");
+
   StoreInst *createAlignedStore(Value *val, Value *ptr, unsigned align);
   // moves the insertion point to the start of the function you're inside,
   // invokes functor, then moves the insertion point back to its original
@@ -280,7 +280,6 @@ public:
   // both branches here:
   // BEGIN { if (nsecs > 0) { $a = 1 } else { $a = 2 } print($a); exit() }
   void hoist(const std::function<void()> &functor);
-  int helper_error_id_ = 0;
 
   // Returns the integer type used to represent pointers in traced code.
   llvm::Type *getPointerStorageTy(AddrSpace as);
@@ -288,7 +287,15 @@ public:
 private:
   Module &module_;
   BPFtrace &bpftrace_;
+  AsyncIds &async_ids_;
 
+  CallInst *CreateGetPidTgid(const location &loc);
+  void CreateGetNsPidTgid(Value *ctx,
+                          Value *dev,
+                          Value *ino,
+                          AllocaInst *ret,
+                          const location &loc);
+  llvm::Type *BpfPidnsInfoType();
   Value *CreateUSDTReadArgument(Value *ctx,
                                 struct bcc_usdt_argument *argument,
                                 Builtin &builtin,
@@ -297,10 +304,6 @@ private:
   CallInst *createMapLookup(const std::string &map_name,
                             Value *key,
                             const std::string &name = "lookup_elem");
-  CallInst *createMapLookup(const std::string &map_name,
-                            Value *key,
-                            PointerType *val_ptr_ty,
-                            const std::string &name = "lookup_elem");
   CallInst *createPerCpuMapLookup(
       const std::string &map_name,
       Value *key,
@@ -314,10 +317,22 @@ private:
       const std::string &name = "lookup_percpu_elem");
   CallInst *createGetScratchMap(const std::string &map_name,
                                 const std::string &name,
-                                PointerType *val_ptr_ty,
                                 const location &loc,
                                 BasicBlock *failure_callback,
                                 int key = 0);
+  Value *CreateReadMapValueAllocation(const SizedType &value_type,
+                                      const std::string &name,
+                                      const location &loc);
+  Value *createAllocation(globalvars::GlobalVar globalvar,
+                          llvm::Type *obj_type,
+                          const std::string &name,
+                          const location &loc,
+                          std::optional<std::function<size_t(AsyncIds &)>>
+                              gen_async_id_cb = std::nullopt);
+  void CreateAllocationInit(const SizedType &stype, Value *alloc);
+  Value *createScratchBuffer(globalvars::GlobalVar globalvar,
+                             const location &loc,
+                             size_t key);
   libbpf::bpf_func_id selectProbeReadHelper(AddrSpace as, bool str);
 
   llvm::Type *getKernelPointerStorageTy();
@@ -335,8 +350,13 @@ private:
                           AllocaInst *is_ret_set,
                           CallInst *call,
                           const SizedType &type);
+  void createPerCpuAvg(AllocaInst *total,
+                       AllocaInst *count,
+                       CallInst *call,
+                       const SizedType &type);
 
   std::map<std::string, StructType *> structs_;
+  llvm::Function *preserve_static_offset_ = nullptr;
 };
 
 } // namespace ast
