@@ -1,11 +1,15 @@
 #pragma once
 
+#include <time.h>
+
 #include <cstdint>
 #include <iostream>
 #include <map>
 #include <memory>
 #include <optional>
 #include <set>
+#include <string_view>
+#include <sys/stat.h>
 #include <unordered_map>
 #include <utility>
 #include <vector>
@@ -19,6 +23,8 @@
 #include "child.h"
 #include "config.h"
 #include "dwarf_parser.h"
+#include "functions.h"
+#include "ksyms.h"
 #include "output.h"
 #include "pcap_writer.h"
 #include "printf.h"
@@ -27,6 +33,7 @@
 #include "required_resources.h"
 #include "struct.h"
 #include "types.h"
+#include "usyms.h"
 #include "utils.h"
 
 #include <bcc/bcc_syms.h>
@@ -35,45 +42,38 @@ namespace bpftrace {
 
 const int timeout_ms = 100;
 
-struct symbol {
-  std::string name;
-  uint64_t start;
-  uint64_t size;
-  uint64_t address;
-};
-
 struct stack_key {
   int64_t stackid;
   uint32_t nr_stack_frames;
 };
 
-enum class DebugLevel;
+enum class DebugStage;
 
 // globals
-extern DebugLevel bt_debug;
+extern std::set<DebugStage> bt_debug;
 extern bool bt_quiet;
 extern bool bt_verbose;
+extern bool dry_run;
 
-enum class DebugLevel { kNone, kDebug, kFullDebug };
+enum class DebugStage {
+  Ast,
+  Codegen,
+  CodegenOpt,
+  Disassemble,
+  Libbpf,
+  Verifier
+};
 
-inline DebugLevel operator++(DebugLevel &level, int)
-{
-  switch (level) {
-    case DebugLevel::kNone:
-      level = DebugLevel::kDebug;
-      break;
-    case DebugLevel::kDebug:
-      level = DebugLevel::kFullDebug;
-      break;
-    case DebugLevel::kFullDebug:
-      // NOTE (mmarchini): should be handled by the caller
-      level = DebugLevel::kNone;
-      break;
-    default:
-      break;
-  }
-  return level;
-}
+const std::unordered_map<std::string_view, DebugStage> debug_stages = {
+  { "ast", DebugStage::Ast },
+  { "codegen", DebugStage::Codegen },
+  { "codegen-opt", DebugStage::CodegenOpt },
+#ifndef NDEBUG
+  { "dis", DebugStage::Disassemble },
+#endif
+  { "libbpf", DebugStage::Libbpf },
+  { "verifier", DebugStage::Verifier },
+};
 
 class WildcardException : public std::exception {
 public:
@@ -99,11 +99,15 @@ public:
         feature_(std::make_unique<BPFfeature>(no_feature)),
         probe_matcher_(std::make_unique<ProbeMatcher>(this)),
         ncpus_(get_possible_cpus().size()),
-        config_(config)
+        max_cpu_id_(get_max_cpu_id()),
+        config_(config),
+        ksyms_(config_),
+        usyms_(config_)
   {
   }
   virtual ~BPFtrace();
-  virtual int add_probe(const ast::AttachPoint &ap,
+  virtual int add_probe(ast::ASTContext &ctx,
+                        const ast::AttachPoint &ap,
                         const ast::Probe &p,
                         int usdt_location_idx = 0);
   Probe generateWatchpointSetupProbe(const ast::AttachPoint &ap,
@@ -121,16 +125,16 @@ public:
   int print_map(const BpfMap &map, uint32_t top, uint32_t div);
   std::string get_stack(int64_t stackid,
                         uint32_t nr_stack_frames,
-                        int pid,
-                        int probe_id,
+                        int32_t pid,
+                        int32_t probe_id,
                         bool ustack,
                         StackType stack_type,
                         int indent = 0);
-  std::string resolve_buf(char *buf, size_t size);
+  std::string resolve_buf(const char *buf, size_t size);
   std::string resolve_ksym(uint64_t addr, bool show_offset = false);
   std::string resolve_usym(uint64_t addr,
-                           int pid,
-                           int probe_id,
+                           int32_t pid,
+                           int32_t probe_id,
                            bool show_offset = false,
                            bool show_module = false);
   std::string resolve_inet(int af, const uint8_t *inet) const;
@@ -158,21 +162,25 @@ public:
   std::optional<int64_t> get_int_literal(const ast::Expression *expr) const;
   std::optional<std::string> get_watchpoint_binary_path() const;
   virtual bool is_traceable_func(const std::string &func_name) const;
-  virtual int get_num_possible_cpus() const;
   virtual std::unordered_set<std::string> get_func_modules(
       const std::string &func_name) const;
-  int create_pcaps(void);
-  void close_pcaps(void);
+  virtual const struct stat &get_pidns_self_stat() const;
+
   bool write_pcaps(uint64_t id, uint64_t ns, uint8_t *pkt, unsigned int size);
 
   void parse_btf(const std::set<std::string> &modules);
   bool has_btf_data() const;
   Dwarf *get_dwarf(const std::string &filename);
   Dwarf *get_dwarf(const ast::AttachPoint &attachpoint);
-  void kfunc_recursion_check(ast::Program *prog);
+  bool has_dwarf_data() const
+  {
+    return !dwarves_.empty();
+  }
+  void fentry_recursion_check(ast::Program *prog);
 
   std::string cmd_;
   bool finalize_ = false;
+  static int exit_code;
   // Global variables checking if an exit/usr1 signal was received
   static volatile sig_atomic_t exitsig_recv;
   static volatile sig_atomic_t sigusr1_recv;
@@ -180,14 +188,17 @@ public:
   RequiredResources resources;
   BpfBytecode bytecode_;
   StructManager structs;
+  FunctionRegistry functions;
   std::map<std::string, std::string> macros_;
-  std::map<std::string, uint64_t> enums_;
+  // Map of enum variant_name to (variant_value, enum_name)
+  std::map<std::string, std::tuple<uint64_t, std::string>> enums_;
+  // Map of enum_name to map of variant_value to variant_name
+  std::map<std::string, std::map<uint64_t, std::string>> enum_defs_;
   std::map<libbpf::bpf_func_id, location> helper_use_loc_;
   const FuncsModulesMap &get_traceable_funcs() const;
   KConfig kconfig;
   std::vector<std::unique_ptr<AttachedProbe>> attached_probes_;
-
-  std::map<std::string, std::unique_ptr<PCAPwriter>> pcap_writers;
+  std::optional<int> sigusr1_prog_fd_;
 
   unsigned int join_argnum_ = 16;
   unsigned int join_argsize_ = 1024;
@@ -198,8 +209,8 @@ public:
   bool safe_mode_ = true;
   bool has_usdt_ = false;
   bool usdt_file_activation_ = false;
-  int helper_check_level_ = 0;
-  uint64_t max_ast_nodes_ = 0; // Maximum AST nodes allowed for fuzzing
+  int helper_check_level_ = 1;
+  uint64_t max_ast_nodes_ = std::numeric_limits<uint64_t>::max();
   bool debug_output_ = false;
   std::optional<struct timespec> boottime_;
   std::optional<struct timespec> delta_taitime_;
@@ -208,7 +219,7 @@ public:
   bool need_recursion_check_ = false;
 
   static void sort_by_key(
-      std::vector<SizedType> key_args,
+      const SizedType &key,
       std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
           &values_by_key);
 
@@ -224,27 +235,24 @@ public:
   }
   int ncpus_;
   int online_cpus_;
+  int max_cpu_id_;
   Config config_;
 
 private:
-  int run_special_probe(std::string name,
-                        const BpfBytecode &bytecode,
-                        void (*trigger)(void));
-  void *ksyms_{ nullptr };
-  // note: exe_sym_ is used when layout is same for all instances of program
-  std::map<std::string, std::pair<int, void *>> exe_sym_; // exe -> (pid, cache)
-  std::map<int, void *> pid_sym_;                         // pid -> cache
-  std::map<std::string, std::map<uintptr_t, elf_symbol, std::greater<>>>
-      symbol_table_cache_;
+  Ksyms ksyms_;
+  Usyms usyms_;
   std::vector<std::string> params_;
 
   std::vector<std::unique_ptr<void, void (*)(void *)>> open_perf_buffers_;
+  std::map<std::string, std::unique_ptr<PCAPwriter>> pcap_writers_;
 
   std::vector<std::unique_ptr<AttachedProbe>> attach_usdt_probe(
       Probe &probe,
-      BpfProgram &&program,
+      const BpfProgram &program,
       int pid,
       bool file_activation);
+  int create_pcaps();
+  void close_pcaps();
   int setup_output();
   int setup_perf_events();
   void setup_ringbuf();
@@ -266,9 +274,7 @@ private:
   int poll_perf_events();
   void handle_event_loss();
   int print_map_hist(const BpfMap &map, uint32_t top, uint32_t div);
-  int print_map_stats(const BpfMap &map, uint32_t top, uint32_t div);
   static uint64_t read_address_from_output(std::string output);
-  std::optional<std::vector<uint8_t>> find_empty_key(const BpfMap &map) const;
   struct bcc_symbol_option &get_symbol_opts();
   Probe generate_probe(const ast::AttachPoint &ap,
                        const ast::Probe &p,
