@@ -1,8 +1,7 @@
 #include "debugfs.h"
 #include <unistd.h>
 
-namespace bpftrace {
-namespace debugfs {
+namespace bpftrace::debugfs {
 
 #define DEBUGFS "/sys/kernel/debug"
 
@@ -16,5 +15,4 @@ std::string path(const std::string &file)
   return path() + "/" + file;
 }
 
-} // namespace debugfs
-} // namespace bpftrace
+} // namespace bpftrace::debugfs
