@@ -40,10 +40,11 @@ enum class ArgumentType {
 
 class IPrintable {
 public:
-  virtual ~IPrintable(){};
+  virtual ~IPrintable() {};
   virtual int print(char* buf,
                     size_t size,
                     const char* fmt,
+                    Type token,
                     ArgumentType expected_type = ArgumentType::UNKNOWN) = 0;
 };
 
@@ -52,7 +53,11 @@ public:
   PrintableString(std::string value,
                   std::optional<size_t> buffer_size = std::nullopt,
                   const char* trunc_trailer = nullptr);
-  int print(char* buf, size_t size, const char* fmt, ArgumentType) override;
+  int print(char* buf,
+            size_t size,
+            const char* fmt,
+            Type,
+            ArgumentType) override;
 
 private:
   std::string value_;
@@ -64,7 +69,11 @@ public:
       : value_(std::vector<char>(buffer, buffer + size))
   {
   }
-  int print(char* buf, size_t size, const char* fmt, ArgumentType) override;
+  int print(char* buf,
+            size_t size,
+            const char* fmt,
+            Type,
+            ArgumentType) override;
   void keep_ascii(bool value);
   void escape_hex(bool value);
 
@@ -79,7 +88,11 @@ public:
   PrintableCString(char* value) : value_(value)
   {
   }
-  int print(char* buf, size_t size, const char* fmt, ArgumentType) override;
+  int print(char* buf,
+            size_t size,
+            const char* fmt,
+            Type,
+            ArgumentType) override;
 
 private:
   char* value_;
@@ -93,6 +106,7 @@ public:
   int print(char* buf,
             size_t size,
             const char* fmt,
+            Type token,
             ArgumentType expected_type) override;
 
 private:
@@ -107,10 +121,27 @@ public:
   int print(char* buf,
             size_t size,
             const char* fmt,
+            Type token,
             ArgumentType expected_type) override;
 
 private:
   int64_t value_;
 };
 
+class PrintableEnum : public virtual IPrintable {
+public:
+  PrintableEnum(uint64_t value, std::string name) : name_(name), value_(value)
+  {
+  }
+  int print(char* buf,
+            size_t size,
+            const char* fmt,
+            Type token,
+            ArgumentType expected_type) override;
+
+private:
+  std::string name_;
+  uint64_t value_;
+};
+
 } // namespace bpftrace
