@@ -1,3 +1,4 @@
+#include <cstdint>
 #include <cstring>
 
 #include "ast/passes/codegen_llvm.h"
@@ -11,9 +12,7 @@
 #include "gmock/gmock.h"
 #include "gtest/gtest.h"
 
-namespace bpftrace {
-namespace test {
-namespace bpftrace {
+namespace bpftrace::test::bpftrace {
 
 #include "btf_common.h"
 
@@ -40,19 +39,18 @@ static auto parse_probe(const std::string &str,
   Driver driver(bpftrace);
   ASSERT_EQ(driver.parse_str(str), 0);
 
-  ast::FieldAnalyser fields(driver.root.get(), bpftrace);
+  ast::FieldAnalyser fields(driver.ctx, bpftrace);
   ASSERT_EQ(fields.analyse(), 0);
 
   ClangParser clang;
-  clang.parse(driver.root.get(), bpftrace);
+  clang.parse(driver.ctx.root, bpftrace);
 
-  ast::SemanticAnalyser semantics(driver.root.get(), bpftrace);
+  ast::SemanticAnalyser semantics(driver.ctx, bpftrace);
   ASSERT_EQ(semantics.analyse(), 0);
 
   auto usdt_helper = get_mock_usdt_helper(usdt_num_locations);
   std::stringstream out;
-  ast::CodegenLLVM codegen(
-      driver.root.get(), bpftrace, false, std::move(usdt_helper));
+  ast::CodegenLLVM codegen(driver.ctx, bpftrace, std::move(usdt_helper));
   codegen.generate_ir();
 }
 
@@ -193,12 +191,9 @@ void check_hardware(Probe &p,
   EXPECT_EQ("hardware:" + unit + ":" + std::to_string(freq), p.name);
 }
 
-void check_special_probe(Probe &p,
-                         const std::string &attach_point,
-                         const std::string &orig_name)
+void check_special_probe(Probe &p, const std::string &orig_name)
 {
   EXPECT_EQ(ProbeType::special, p.type);
-  EXPECT_EQ(attach_point, p.attach_point);
   EXPECT_EQ(orig_name, p.orig_name);
   EXPECT_EQ(orig_name, p.name);
 }
@@ -211,9 +206,7 @@ TEST(bpftrace, add_begin_probe)
   ASSERT_EQ(0U, bpftrace.get_probes().size());
   ASSERT_EQ(1U, bpftrace.get_special_probes().size());
 
-  check_special_probe(bpftrace.get_special_probes().at(0),
-                      "BEGIN_trigger",
-                      "BEGIN");
+  check_special_probe(bpftrace.get_special_probes()["BEGIN"], "BEGIN");
 }
 
 TEST(bpftrace, add_end_probe)
@@ -224,9 +217,7 @@ TEST(bpftrace, add_end_probe)
   ASSERT_EQ(0U, bpftrace.get_probes().size());
   ASSERT_EQ(1U, bpftrace.get_special_probes().size());
 
-  check_special_probe(bpftrace.get_special_probes().at(0),
-                      "END_trigger",
-                      "END");
+  check_special_probe(bpftrace.get_special_probes()["END"], "END");
 }
 
 TEST(bpftrace, add_probes_single)
@@ -274,10 +265,9 @@ TEST(bpftrace, add_probes_wildcard)
 TEST(bpftrace, add_probes_wildcard_kprobe_multi)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_symbols_from_traceable_funcs(false))
-      .Times(1);
+      .Times(2);
 
   parse_probe("kprobe:sys_read,kprobe:my_*,kprobe:sys_write{}", *bpftrace);
 
@@ -316,7 +306,7 @@ TEST(bpftrace, add_probes_wildcard_no_matches)
 TEST(bpftrace, add_probes_wildcard_no_matches_kprobe_multi)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_symbols_from_traceable_funcs(false))
       .Times(1);
@@ -366,7 +356,7 @@ TEST(bpftrace, add_probes_kernel_module_wildcard)
   auto bpftrace = get_strict_mock_bpftrace();
   // We enable kprobe_multi here but it doesn't support the module:function
   // syntax so full expansion should be done anyways.
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_symbols_from_traceable_funcs(true))
       .Times(1);
@@ -399,7 +389,7 @@ TEST(bpftrace, add_probes_kernel_module_function_wildcard)
   auto bpftrace = get_strict_mock_bpftrace();
   // We enable kprobe_multi here but it doesn't support the module:function
   // syntax so full expansion should be done anyways.
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_symbols_from_traceable_funcs(true))
       .Times(1);
@@ -478,10 +468,10 @@ TEST(bpftrace, add_probes_uprobe_wildcard)
 TEST(bpftrace, add_probes_uprobe_wildcard_uprobe_multi)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_func_symbols_from_file(0, "/bin/sh"))
-      .Times(1);
+      .Times(2);
 
   parse_probe("uprobe:/bin/sh:*open {}", *bpftrace);
 
@@ -530,10 +520,10 @@ TEST(bpftrace, add_probes_uprobe_wildcard_file)
 TEST(bpftrace, add_probes_uprobe_wildcard_file_uprobe_multi)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_func_symbols_from_file(0, "/bin/*sh"))
-      .Times(1);
+      .Times(2);
 
   parse_probe("uprobe:/bin/*sh:*open {}", *bpftrace);
 
@@ -570,7 +560,7 @@ TEST(bpftrace, add_probes_uprobe_wildcard_no_matches)
 TEST(bpftrace, add_probes_uprobe_wildcard_no_matches_multi)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   EXPECT_CALL(*bpftrace->mock_probe_matcher,
               get_func_symbols_from_file(0, "/bin/sh"))
       .Times(1);
@@ -717,6 +707,46 @@ TEST(bpftrace, add_probes_uprobe_no_demangling)
                "uprobe:/bin/sh:cpp_mangled");
 }
 
+#ifdef HAVE_LIBLLDB
+#include "dwarf_common.h"
+
+class bpftrace_dwarf : public test_dwarf {};
+
+TEST_F(bpftrace_dwarf, add_probes_uprobe_symbol_source)
+{
+  auto uprobe = "uprobe:" + std::string(bin_) + ":func_1 {}";
+
+  {
+    BPFtrace bpftrace;
+    ConfigSetter configs{ bpftrace.config_, ConfigSource::script };
+    configs.set_symbol_source_config("dwarf");
+    parse_probe(uprobe, bpftrace);
+
+    ASSERT_EQ(bpftrace.resources.probes.size(), 1);
+    ASSERT_EQ(bpftrace.resources.special_probes.size(), 0);
+
+    auto &probe = bpftrace.resources.probes.at(0);
+    ASSERT_TRUE(probe.attach_point.empty());
+    ASSERT_NE(probe.address, 0);
+  }
+
+  {
+    BPFtrace bpftrace;
+    ConfigSetter configs{ bpftrace.config_, ConfigSource::script };
+    configs.set_symbol_source_config("symbol_table");
+    parse_probe(uprobe, bpftrace);
+
+    ASSERT_EQ(bpftrace.resources.probes.size(), 1);
+    ASSERT_EQ(bpftrace.resources.special_probes.size(), 0);
+
+    auto &probe = bpftrace.resources.probes.at(0);
+    ASSERT_FALSE(probe.attach_point.empty());
+    ASSERT_EQ(probe.address, 0);
+  }
+}
+
+#endif
+
 TEST(bpftrace, add_probes_usdt)
 {
   StrictMock<MockBPFtrace> bpftrace;
@@ -969,7 +999,9 @@ std::pair<std::vector<uint8_t>, std::vector<uint8_t>> key_value_pair_str(
   uint8_t *val_data = pair.second.data();
 
   for (size_t i = 0; i < key.size(); i++) {
-    strncpy((char *)key_data + STRING_SIZE * i, key.at(i).c_str(), STRING_SIZE);
+    strncpy(reinterpret_cast<char *>(key_data) + STRING_SIZE * i,
+            key.at(i).c_str(),
+            STRING_SIZE);
   }
   uint64_t v = val;
   std::memcpy(val_data, &v, sizeof(v));
@@ -991,7 +1023,9 @@ std::pair<std::vector<uint8_t>, std::vector<uint8_t>> key_value_pair_int_str(
 
   uint64_t k = myint, v = val;
   std::memcpy(key_data, &k, sizeof(k));
-  strncpy((char *)key_data + sizeof(uint64_t), mystr.c_str(), STRING_SIZE);
+  strncpy(reinterpret_cast<char *>(key_data) + sizeof(uint64_t),
+          mystr.c_str(),
+          STRING_SIZE);
   std::memcpy(val_data, &v, sizeof(v));
 
   return pair;
@@ -1001,14 +1035,14 @@ TEST(bpftrace, sort_by_key_int)
 {
   StrictMock<MockBPFtrace> bpftrace;
 
-  std::vector<SizedType> key_args = { CreateUInt64() };
+  SizedType key_arg = CreateUInt64();
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       values_by_key = {
         key_value_pair_int({ 2 }, 12),
         key_value_pair_int({ 3 }, 11),
         key_value_pair_int({ 1 }, 10),
       };
-  bpftrace.sort_by_key(key_args, values_by_key);
+  bpftrace.sort_by_key(key_arg, values_by_key);
 
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       expected_values = {
@@ -1024,18 +1058,16 @@ TEST(bpftrace, sort_by_key_int_int)
 {
   StrictMock<MockBPFtrace> bpftrace;
 
-  std::vector<SizedType> key_args = {
-    CreateUInt64(),
-    CreateUInt64(),
-    CreateUInt64(),
-  };
+  SizedType key = CreateTuple(bpftrace.structs.AddTuple(
+      { CreateInt64(), CreateInt64(), CreateInt64() }));
+
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       values_by_key = {
         key_value_pair_int({ 5, 2, 1 }, 1), key_value_pair_int({ 5, 3, 1 }, 2),
         key_value_pair_int({ 5, 1, 1 }, 3), key_value_pair_int({ 2, 2, 2 }, 4),
         key_value_pair_int({ 2, 3, 2 }, 5), key_value_pair_int({ 2, 1, 2 }, 6),
       };
-  bpftrace.sort_by_key(key_args, values_by_key);
+  bpftrace.sort_by_key(key, values_by_key);
 
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       expected_values = {
@@ -1051,9 +1083,7 @@ TEST(bpftrace, sort_by_key_str)
 {
   StrictMock<MockBPFtrace> bpftrace;
 
-  std::vector<SizedType> key_args = {
-    CreateString(STRING_SIZE),
-  };
+  SizedType key_arg = CreateString(STRING_SIZE);
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       values_by_key = {
         key_value_pair_str({ "z" }, 1),
@@ -1061,7 +1091,7 @@ TEST(bpftrace, sort_by_key_str)
         key_value_pair_str({ "x" }, 3),
         key_value_pair_str({ "d" }, 4),
       };
-  bpftrace.sort_by_key(key_args, values_by_key);
+  bpftrace.sort_by_key(key_arg, values_by_key);
 
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       expected_values = {
@@ -1078,11 +1108,11 @@ TEST(bpftrace, sort_by_key_str_str)
 {
   StrictMock<MockBPFtrace> bpftrace;
 
-  std::vector<SizedType> key_args = {
-    CreateString(STRING_SIZE),
-    CreateString(STRING_SIZE),
-    CreateString(STRING_SIZE),
-  };
+  SizedType key = CreateTuple(
+      bpftrace.structs.AddTuple({ CreateString(STRING_SIZE),
+                                  CreateString(STRING_SIZE),
+                                  CreateString(STRING_SIZE) }));
+
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       values_by_key = {
         key_value_pair_str({ "z", "a", "l" }, 1),
@@ -1092,7 +1122,7 @@ TEST(bpftrace, sort_by_key_str_str)
         key_value_pair_str({ "z", "b", "p" }, 5),
         key_value_pair_str({ "a", "b", "q" }, 6),
       };
-  bpftrace.sort_by_key(key_args, values_by_key);
+  bpftrace.sort_by_key(key, values_by_key);
 
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       expected_values = {
@@ -1111,17 +1141,16 @@ TEST(bpftrace, sort_by_key_int_str)
 {
   StrictMock<MockBPFtrace> bpftrace;
 
-  std::vector<SizedType> key_args = {
-    CreateUInt64(),
-    CreateString(STRING_SIZE),
-  };
+  SizedType key = CreateTuple(
+      bpftrace.structs.AddTuple({ CreateUInt64(), CreateString(STRING_SIZE) }));
+
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       values_by_key = {
         key_value_pair_int_str(1, "b", 1), key_value_pair_int_str(2, "b", 2),
         key_value_pair_int_str(3, "b", 3), key_value_pair_int_str(1, "a", 4),
         key_value_pair_int_str(2, "a", 5), key_value_pair_int_str(3, "a", 6),
       };
-  bpftrace.sort_by_key(key_args, values_by_key);
+  bpftrace.sort_by_key(key, values_by_key);
 
   std::vector<std::pair<std::vector<uint8_t>, std::vector<uint8_t>>>
       expected_values = {
@@ -1141,21 +1170,21 @@ void check_probe(Probe &p, ProbeType type, const std::string &name)
   EXPECT_EQ(name, p.name);
 }
 
-TEST_F(bpftrace_btf, add_probes_kfunc)
+TEST_F(bpftrace_btf, add_probes_fentry)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
-  parse_probe("kfunc:func_1,kretfunc:func_1 {}", *bpftrace);
+
+  parse_probe("fentry:func_1,fexit:func_1 {}", *bpftrace);
 
   ASSERT_EQ(2U, bpftrace->get_probes().size());
   ASSERT_EQ(0U, bpftrace->get_special_probes().size());
 
   check_probe(bpftrace->get_probes().at(0),
-              ProbeType::kfunc,
-              "kfunc:mock_vmlinux:func_1");
+              ProbeType::fentry,
+              "fentry:mock_vmlinux:func_1");
   check_probe(bpftrace->get_probes().at(1),
-              ProbeType::kretfunc,
-              "kretfunc:mock_vmlinux:func_1");
+              ProbeType::fexit,
+              "fexit:mock_vmlinux:func_1");
 }
 
 TEST_F(bpftrace_btf, add_probes_kprobe)
@@ -1178,7 +1207,7 @@ TEST_F(bpftrace_btf, add_probes_kprobe)
 TEST_F(bpftrace_btf, add_probes_iter_task)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   parse_probe("iter:task {}", *bpftrace);
 
   ASSERT_EQ(1U, bpftrace->get_probes().size());
@@ -1190,7 +1219,7 @@ TEST_F(bpftrace_btf, add_probes_iter_task)
 TEST_F(bpftrace_btf, add_probes_iter_task_file)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   parse_probe("iter:task_file {}", *bpftrace);
 
   ASSERT_EQ(1U, bpftrace->get_probes().size());
@@ -1202,7 +1231,7 @@ TEST_F(bpftrace_btf, add_probes_iter_task_file)
 TEST_F(bpftrace_btf, add_probes_iter_task_vma)
 {
   auto bpftrace = get_strict_mock_bpftrace();
-  bpftrace->feature_ = std::make_unique<MockBPFfeature>(true);
+
   parse_probe("iter:task_vma {}", *bpftrace);
 
   ASSERT_EQ(1U, bpftrace->get_probes().size());
@@ -1211,6 +1240,16 @@ TEST_F(bpftrace_btf, add_probes_iter_task_vma)
   check_probe(bpftrace->get_probes().at(0), ProbeType::iter, "iter:task_vma");
 }
 
+class bpftrace_bad_btf : public test_bad_btf {};
+
+// Test that we can handle bad data and don't just crash
+TEST_F(bpftrace_bad_btf, parse_invalid_btf)
+{
+  BPFtrace bpftrace;
+  bpftrace.parse_btf({});
+  EXPECT_FALSE(bpftrace.has_btf_data());
+}
+
 TEST(bpftrace, add_probes_rawtracepoint)
 {
   StrictMock<MockBPFtrace> bpftrace;
@@ -1251,6 +1290,27 @@ TEST(bpftrace, add_probes_rawtracepoint_wildcard_no_matches)
   ASSERT_EQ(0U, bpftrace->get_special_probes().size());
 }
 
-} // namespace bpftrace
-} // namespace test
-} // namespace bpftrace
+TEST(bpftrace, resolve_timestamp)
+{
+  static const auto bootmode = static_cast<uint32_t>(TimestampMode::boot);
+  auto bpftrace = get_strict_mock_bpftrace();
+
+  // Basic sanity check
+  bpftrace->boottime_ = { .tv_sec = 3, .tv_nsec = 0 };
+  bpftrace->resources.strftime_args.push_back("%s.%f");
+  EXPECT_EQ(bpftrace->resolve_timestamp(bootmode, 0, 1000), "3.000001");
+
+  // Check that boottime nsecs close to 1s doesn't trigger floating-point error.
+  //
+  // Due to the peculiarities of floating-point, not _any_ set of numbers
+  // trigger the bug here. These values were discovered in the wild.
+  bpftrace->boottime_ = { .tv_sec = 1736725826, .tv_nsec = 999999985 };
+  bpftrace->resources.strftime_args.push_back("%s");
+  EXPECT_EQ(bpftrace->resolve_timestamp(bootmode, 1, 0), "1736725826");
+
+  // Now check that we handle rollover to a new second correctly
+  bpftrace->resources.strftime_args.push_back("%s.%f");
+  EXPECT_EQ(bpftrace->resolve_timestamp(bootmode, 2, 15), "1736725827.000000");
+}
+
+} // namespace bpftrace::test::bpftrace
