@@ -9,7 +9,7 @@ TEST(codegen, struct_string_array)
   test("struct Foo { char str[32]; }"
        "kprobe:f"
        "{"
-       "  $foo = (struct Foo)0;"
+       "  $foo = *(struct Foo*)arg0;"
        "  @mystr = $foo.str;"
        "}",
        std::string(NAME) + "_1");
@@ -17,7 +17,7 @@ TEST(codegen, struct_string_array)
   test("struct Foo { char str[32]; }"
        "kprobe:f"
        "{"
-       "  $foo = (struct Foo*)0;"
+       "  $foo = (struct Foo*)arg0;"
        "  @mystr = $foo->str;"
        "}",
        std::string(NAME) + "_2");
