@@ -4,8 +4,7 @@
 #include "gtest/gtest.h"
 #include <iostream>
 
-namespace bpftrace {
-namespace test {
+namespace bpftrace::test {
 
 TEST(Config, get_and_set)
 {
@@ -49,6 +48,11 @@ TEST(Config, get_and_set)
   EXPECT_TRUE(config_setter.set(StackMode::bpftrace));
   EXPECT_EQ(config.get(ConfigKeyStackMode::default_), StackMode::bpftrace);
 
+  // Test that this is also true by default, as a requirement.
+  EXPECT_TRUE(config.get(ConfigKeyBool::print_maps_on_exit));
+  EXPECT_TRUE(config_setter.set(ConfigKeyBool::print_maps_on_exit, false));
+  EXPECT_EQ(config.get(ConfigKeyBool::print_maps_on_exit), false);
+
   EXPECT_TRUE(config_setter.set(UserSymbolCacheType::per_program));
   EXPECT_EQ(config.get(ConfigKeyUserSymbolCacheType::default_),
             UserSymbolCacheType::per_program);
@@ -134,5 +138,4 @@ TEST(ConfigSetter, same_source_cannot_set_twice)
   EXPECT_FALSE(config_setter.set(ConfigKeyInt::max_map_keys, 11));
 }
 
-} // namespace test
-} // namespace bpftrace
+} // namespace bpftrace::test
