@@ -75,6 +75,24 @@ public:
     return { "mock_vmlinux" };
   }
 
+  const struct stat &get_pidns_self_stat() const override
+  {
+    static const struct stat init_pid_namespace = []() {
+      struct stat s {};
+      s.st_ino = 0xeffffffc; // PROC_PID_INIT_INO
+      return s;
+    }();
+    static const struct stat child_pid_namespace = []() {
+      struct stat s {};
+      s.st_ino = 0xf0000011; // Arbitrary user namespace
+      return s;
+    }();
+
+    if (mock_in_init_pid_ns)
+      return init_pid_namespace;
+    return child_pid_namespace;
+  }
+
   void set_mock_probe_matcher(std::unique_ptr<MockProbeMatcher> probe_matcher)
   {
     probe_matcher_ = std::move(probe_matcher);
@@ -82,6 +100,7 @@ public:
   }
 
   MockProbeMatcher *mock_probe_matcher;
+  bool mock_in_init_pid_ns = true;
 };
 
 std::unique_ptr<MockBPFtrace> get_mock_bpftrace();
@@ -95,7 +114,6 @@ public:
     has_get_current_cgroup_id_ = std::make_optional<bool>(has_features);
     has_override_return_ = std::make_optional<bool>(has_features);
     has_prog_fentry_ = std::make_optional<bool>(has_features);
-    has_loop_ = std::make_optional<bool>(has_features);
     has_probe_read_kernel_ = std::make_optional<bool>(has_features);
     has_features_ = has_features;
     has_d_path_ = std::make_optional<bool>(has_features);
@@ -108,13 +126,10 @@ public:
     has_get_func_ip_ = std::make_optional<bool>(has_features);
     has_jiffies64_ = std::make_optional<bool>(has_features);
     has_for_each_map_elem_ = std::make_optional<bool>(has_features);
+    has_get_ns_current_pid_tgid_ = std::make_optional<bool>(has_features);
+    has_map_lookup_percpu_elem_ = std::make_optional<bool>(has_features);
   };
 
-  void has_loop(bool has)
-  {
-    has_loop_ = std::make_optional<bool>(has);
-  }
-
   void mock_missing_kernel_func(Kfunc kfunc)
   {
     available_kernel_funcs_.emplace(kfunc, false);
@@ -129,14 +144,14 @@ public:
   {
     child_pid_ = 1337;
   };
-  ~MockChildProc(){};
+  ~MockChildProc() {};
 
-  void terminate(bool force __attribute__((unused)) = false) override{};
+  void terminate(bool force __attribute__((unused)) = false) override {};
   bool is_alive() override
   {
     return true;
   };
-  void resume(void) override{};
+  void resume(void) override {};
 
   void run(bool pause = false) override
   {
