@@ -1,13 +1,11 @@
-#include <limits.h>
+#include <climits>
 #include <sstream>
 
 #include "ast/passes/printer.h"
 #include "driver.h"
 #include "gtest/gtest.h"
 
-namespace bpftrace {
-namespace test {
-namespace parser {
+namespace bpftrace::test::parser {
 
 using Printer = ast::Printer;
 
@@ -41,8 +39,8 @@ void test(BPFtrace &bpftrace, std::string_view input, std::string_view expected)
     expected.remove_prefix(1); // Remove initial '\n'
 
   std::ostringstream out;
-  Printer printer(out);
-  printer.print(driver.root.get());
+  Printer printer(driver.ctx, out);
+  printer.print();
   EXPECT_EQ(expected, out.str());
 }
 
@@ -286,9 +284,9 @@ TEST(Parser, positional_param_attachpoint)
 stdin:1:1-12: ERROR: Found trailing text 'a' in positional parameter index. Try quoting the trailing text.
 uprobe:$1a { 1 }
 ~~~~~~~~~~~
-stdin:1:1-1: ERROR: No attach points for probe
+stdin:1:1-18: ERROR: No attach points for probe
 uprobe:$1a { 1 }
-
+~~~~~~~~~~~~~~~~
 )");
 
   test_parse_failure(bpftrace, R"(uprobe:$a { 1 })", R"(
@@ -679,9 +677,21 @@ TEST(Parser, map_key)
        "   int: 1\n"
        "  =\n"
        "   map: @x\n"
-       "    int: 0\n"
-       "    int: 1\n"
-       "    int: 2\n"
+       "    tuple:\n"
+       "     int: 0\n"
+       "     int: 1\n"
+       "     int: 2\n"
+       "   int: 1\n");
+
+  test("kprobe:sys_open { @x[(0,\"hi\",tid)] = 1; }",
+       "Program\n"
+       " kprobe:sys_open\n"
+       "  =\n"
+       "   map: @x\n"
+       "    tuple:\n"
+       "     int: 0\n"
+       "     string: hi\n"
+       "     builtin: tid\n"
        "   int: 1\n");
 
   test("kprobe:sys_open { @x[@a] = 1; @x[@a,@b,@c] = 1; }",
@@ -693,9 +703,10 @@ TEST(Parser, map_key)
        "   int: 1\n"
        "  =\n"
        "   map: @x\n"
-       "    map: @a\n"
-       "    map: @b\n"
-       "    map: @c\n"
+       "    tuple:\n"
+       "     map: @a\n"
+       "     map: @b\n"
+       "     map: @c\n"
        "   int: 1\n");
 
   test("kprobe:sys_open { @x[pid] = 1; @x[tid,uid,arg9] = 1; }",
@@ -707,9 +718,10 @@ TEST(Parser, map_key)
        "   int: 1\n"
        "  =\n"
        "   map: @x\n"
-       "    builtin: tid\n"
-       "    builtin: uid\n"
-       "    builtin: arg9\n"
+       "    tuple:\n"
+       "     builtin: tid\n"
+       "     builtin: uid\n"
+       "     builtin: arg9\n"
        "   int: 1\n");
 }
 
@@ -771,22 +783,22 @@ TEST(Parser, variable_post_increment_decrement)
   test("kprobe:sys_open { $x++; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  variable: $x\n"
-       "   ++\n");
+       "  ++ (post)\n"
+       "   variable: $x\n");
   test("kprobe:sys_open { ++$x; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  ++\n"
+       "  ++ (pre)\n"
        "   variable: $x\n");
   test("kprobe:sys_open { $x--; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  variable: $x\n"
-       "   --\n");
+       "  -- (post)\n"
+       "   variable: $x\n");
   test("kprobe:sys_open { --$x; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  --\n"
+       "  -- (pre)\n"
        "   variable: $x\n");
 }
 
@@ -795,22 +807,22 @@ TEST(Parser, map_increment_decrement)
   test("kprobe:sys_open { @x++; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  map: @x\n"
-       "   ++\n");
+       "  ++ (post)\n"
+       "   map: @x\n");
   test("kprobe:sys_open { ++@x; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  ++\n"
+       "  ++ (pre)\n"
        "   map: @x\n");
   test("kprobe:sys_open { @x--; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  map: @x\n"
-       "   --\n");
+       "  -- (post)\n"
+       "   map: @x\n");
   test("kprobe:sys_open { --@x; }",
        "Program\n"
        " kprobe:sys_open\n"
-       "  --\n"
+       "  -- (pre)\n"
        "   map: @x\n");
 }
 
@@ -1059,7 +1071,7 @@ TEST(Parser, unroll)
 
 TEST(Parser, ternary_str)
 {
-  test("kprobe:sys_open { @x = pid < 10000 ? \"lo\" : \"high\" }",
+  test(R"(kprobe:sys_open { @x = pid < 10000 ? "lo" : "high" })",
        "Program\n"
        " kprobe:sys_open\n"
        "  =\n"
@@ -1236,7 +1248,7 @@ TEST(Parser, usdt)
   // Without the escapes needed for C++ to compile:
   //    usdt:/my/program:"\"probe\"" { 1; }
   //
-  test("usdt:/my/program:\"\\\"probe\\\"\" { 1; }",
+  test(R"(usdt:/my/program:"\"probe\"" { 1; })",
        "Program\n"
        " usdt:/my/program:\"probe\"\n"
        "  int: 1\n");
@@ -1796,16 +1808,25 @@ TEST(Parser, cast_simple_type_pointer)
 
 TEST(Parser, cast_sized_type)
 {
-  test("kprobe:sys_read { (str_t[1])arg0; }",
+  test("kprobe:sys_read { (string)arg0; }",
        "Program\n"
        " kprobe:sys_read\n"
-       "  (string[1])\n"
+       "  (string[0])\n"
        "   builtin: arg0\n");
 }
 
 TEST(Parser, cast_sized_type_pointer)
 {
-  test("kprobe:sys_read { (str_t[1] *)arg0; }",
+  test("kprobe:sys_read { (string *)arg0; }",
+       "Program\n"
+       " kprobe:sys_read\n"
+       "  (string[0] *)\n"
+       "   builtin: arg0\n");
+}
+
+TEST(Parser, cast_sized_type_pointer_with_size)
+{
+  test("kprobe:sys_read { (string[1] *)arg0; }",
        "Program\n"
        " kprobe:sys_read\n"
        "  (string[1] *)\n"
@@ -1915,6 +1936,17 @@ TEST(Parser, cast_precedence)
        "   int: 123\n");
 }
 
+TEST(Parser, cast_enum)
+{
+  test("enum Foo { ONE = 1 } kprobe:sys_read { (enum Foo)1; }",
+       "enum Foo { ONE = 1 };\n"
+       "\n"
+       "Program\n"
+       " kprobe:sys_read\n"
+       "  (enum Foo)\n"
+       "   int: 1\n");
+}
+
 TEST(Parser, sizeof_expression)
 {
   test("kprobe:sys_read { sizeof(arg0); }",
@@ -1942,6 +1974,23 @@ TEST(Parser, offsetof_type)
        "  offsetof: \n"
        "   struct Foo\n"
        "   x\n");
+  test("struct Foo { struct Bar { int x; } bar; } "
+       "BEGIN { offsetof(struct Foo, bar.x); }",
+       "struct Foo { struct Bar { int x; } bar; };\n"
+       "\n"
+       "Program\n"
+       " BEGIN\n"
+       "  offsetof: \n"
+       "   struct Foo\n"
+       "   bar\n"
+       "   x\n");
+  test_parse_failure("struct Foo { struct Bar { int x; } *bar; } "
+                     "BEGIN { offsetof(struct Foo, bar->x); }",
+                     R"(
+stdin:1:74-79: ERROR: syntax error, unexpected ->, expecting ) or .
+struct Foo { struct Bar { int x; } *bar; } BEGIN { offsetof(struct Foo, bar->x); }
+                                                                         ~~~~~
+)");
 }
 
 TEST(Parser, offsetof_expression)
@@ -1962,6 +2011,18 @@ TEST(Parser, offsetof_expression)
        "   x\n");
 }
 
+TEST(Parser, offsetof_builtin_type)
+{
+  test("struct Foo { timestamp x; } BEGIN { offsetof(struct Foo, timestamp); }",
+       "struct Foo { timestamp x; };\n"
+       "\n"
+       "Program\n"
+       " BEGIN\n"
+       "  offsetof: \n"
+       "   struct Foo\n"
+       "   timestamp\n");
+}
+
 TEST(Parser, dereference_precedence)
 {
   test("kprobe:sys_read { *@x+1 }",
@@ -2034,6 +2095,24 @@ TEST(Parser, field_access_builtin)
        "   count\n");
 }
 
+TEST(Parser, field_access_builtin_type)
+{
+  test("kprobe:sys_read { @x.timestamp; }",
+       "Program\n"
+       " kprobe:sys_read\n"
+       "  .\n"
+       "   map: @x\n"
+       "   timestamp\n");
+
+  test("kprobe:sys_read { @x->timestamp; }",
+       "Program\n"
+       " kprobe:sys_read\n"
+       "  .\n"
+       "   dereference\n"
+       "    map: @x\n"
+       "   timestamp\n");
+}
+
 TEST(Parser, array_access)
 {
   test("kprobe:sys_read { x[index]; }",
@@ -2090,7 +2169,7 @@ TEST(Parser, unexpected_symbol)
   Driver driver(bpftrace, out);
   EXPECT_EQ(driver.parse_str("i:s:1 { < }"), 1);
   std::string expected =
-      R"(stdin:1:9-10: ERROR: syntax error, unexpected <, expecting }
+      R"(stdin:1:9-10: ERROR: syntax error, unexpected <
 i:s:1 { < }
         ~
 )";
@@ -2121,7 +2200,7 @@ TEST(Parser, unterminated_string)
       R"(stdin:1:12-19: ERROR: unterminated string
 kprobe:f { "asdf }
            ~~~~~~~
-stdin:1:12-19: ERROR: syntax error, unexpected end of file, expecting }
+stdin:1:12-19: ERROR: syntax error, unexpected end of file
 kprobe:f { "asdf }
            ~~~~~~~
 )";
@@ -2197,7 +2276,7 @@ uretprobe:/bin/sh:f+0x10 { 1 }
 TEST(Parser, invalid_increment_decrement)
 {
   test_parse_failure("i:s:1 { @=5++}", R"(
-stdin:1:9-14: ERROR: syntax error, unexpected ++, expecting }
+stdin:1:9-14: ERROR: syntax error, unexpected ++, expecting ; or }
 i:s:1 { @=5++}
         ~~~~~
 )");
@@ -2209,7 +2288,7 @@ i:s:1 { @=++5}
 )");
 
   test_parse_failure("i:s:1 { @=5--}", R"(
-stdin:1:9-14: ERROR: syntax error, unexpected --, expecting }
+stdin:1:9-14: ERROR: syntax error, unexpected --, expecting ; or }
 i:s:1 { @=5--}
         ~~~~~
 )");
@@ -2221,7 +2300,7 @@ i:s:1 { @=--5}
 )");
 
   test_parse_failure("i:s:1 { @=\"a\"++}", R"(
-stdin:1:9-16: ERROR: syntax error, unexpected ++, expecting }
+stdin:1:9-16: ERROR: syntax error, unexpected ++, expecting ; or }
 i:s:1 { @="a"++}
         ~~~~~~~
 )");
@@ -2247,9 +2326,9 @@ TEST(Parser, long_param_overflow)
 TEST(Parser, empty_arguments)
 {
   test_parse_failure("::k::vfs_open:: { 1 }", R"(
-stdin:1:1-1: ERROR: No attach points for probe
+stdin:1:1-26: ERROR: No attach points for probe
 ::k::vfs_open:: { 1 }
-
+~~~~~~~~~~~~~~~~~~~~~
 )");
 
   // Error location is incorrect: #3063
@@ -2260,9 +2339,9 @@ k:vfs_open:: { 1 }
 )");
 
   test_parse_failure(":w:0x10000000:8:rw { 1 }", R"(
-stdin:1:1-1: ERROR: No attach points for probe
+stdin:1:1-25: ERROR: No attach points for probe
 :w:0x10000000:8:rw { 1 }
-
+~~~~~~~~~~~~~~~~~~~~~~~~
 )");
 }
 
@@ -2370,8 +2449,8 @@ TEST(Parser, while_loop)
     variable: $a
     int: 10
    )
-    variable: $a
-     ++
+    ++ (post)
+     variable: $a
 )PROG");
 }
 
@@ -2392,7 +2471,7 @@ i:s:1 { @x = (1, 2); $x.1 = 1; }
 TEST(Parser, tuple_assignment_error)
 {
   test_parse_failure("i:s:1 { (1, 0) = 0 }", R"(
-stdin:1:16-17: ERROR: syntax error, unexpected =, expecting }
+stdin:1:16-17: ERROR: syntax error, unexpected =, expecting ; or }
 i:s:1 { (1, 0) = 0 }
                ~
 )");
@@ -2409,7 +2488,7 @@ i:s:1 { ((1, 0), 3).0 = (0, 1) }
         ~~~~~~~~~~~~~~~~~~~~~~
 )");
 
-  test_parse_failure("i:s:1 { (1, \"two\", (3, 4)).5 = \"six\"; }", R"(
+  test_parse_failure(R"(i:s:1 { (1, "two", (3, 4)).5 = "six"; })", R"(
 stdin:1:9-37: ERROR: Tuples are immutable once created. Consider creating a new tuple and assigning it instead.
 i:s:1 { (1, "two", (3, 4)).5 = "six"; }
         ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -2525,13 +2604,13 @@ i:s:1 { exit(); } config = { BPFTRACE_STACK_MODE=perf }
 )");
 
   test_parse_failure("config = { exit(); } i:s:1 { exit(); }", R"(
-stdin:1:12-16: ERROR: syntax error, unexpected call, expecting }
+stdin:1:12-16: ERROR: syntax error, unexpected call, expecting } or identifier
 config = { exit(); } i:s:1 { exit(); }
            ~~~~
 )");
 
   test_parse_failure("config = { @start = nsecs; } i:s:1 { exit(); }", R"(
-stdin:1:12-18: ERROR: syntax error, unexpected map, expecting }
+stdin:1:12-18: ERROR: syntax error, unexpected map, expecting } or identifier
 config = { @start = nsecs; } i:s:1 { exit(); }
            ~~~~~~
 )");
@@ -2547,7 +2626,7 @@ BEGIN { @start = nsecs; } config = { BPFTRACE_STACK_MODE=perf } i:s:1 { exit();
   test_parse_failure("config = { BPFTRACE_STACK_MODE=perf "
                      "BPFTRACE_MAX_PROBES=2 } i:s:1 { exit(); }",
                      R"(
-stdin:1:37-56: ERROR: syntax error, unexpected identifier, expecting }
+stdin:1:37-56: ERROR: syntax error, unexpected identifier, expecting ; or }
 config = { BPFTRACE_STACK_MODE=perf BPFTRACE_MAX_PROBES=2 } i:s:1 { exit(); }
                                     ~~~~~~~~~~~~~~~~~~~
 )");
@@ -2555,7 +2634,7 @@ config = { BPFTRACE_STACK_MODE=perf BPFTRACE_MAX_PROBES=2 } i:s:1 { exit(); }
   test_parse_failure("config = { BPFTRACE_STACK_MODE=perf } i:s:1 { "
                      "BPFTRACE_MAX_PROBES=2; exit(); }",
                      R"(
-stdin:1:47-67: ERROR: syntax error, unexpected =, expecting }
+stdin:1:47-67: ERROR: syntax error, unexpected =, expecting ; or }
 config = { BPFTRACE_STACK_MODE=perf } i:s:1 { BPFTRACE_MAX_PROBES=2; exit(); }
                                               ~~~~~~~~~~~~~~~~~~~~
 )");
@@ -2628,19 +2707,19 @@ TEST(Parser, subprog_one_arg)
 {
   test("fn f($a : uint8): void {}",
        "Program\n"
-       " f: void($a : unsigned int8)\n");
+       " f: void($a : uint8)\n");
 }
 
 TEST(Parser, subprog_two_args)
 {
   test("fn f($a : uint8, $b : uint8): void {}",
        "Program\n"
-       " f: void($a : unsigned int8, $b : unsigned int8)\n");
+       " f: void($a : uint8, $b : uint8)\n");
 }
 
 TEST(Parser, subprog_string_arg)
 {
-  test("fn f($a : str_t[16]): void {}",
+  test("fn f($a : string[16]): void {}",
        "Program\n"
        " f: void($a : string[16])\n");
 }
@@ -2689,7 +2768,7 @@ TEST(Parser, subprog_return)
 
 TEST(Parser, subprog_string)
 {
-  test("fn f(): str_t[16] {}",
+  test("fn f(): string[16] {}",
        "Program\n"
        " f: string[16]()\n");
 }
@@ -2746,6 +2825,51 @@ BEGIN { for (@kv : @map) { } }
 )");
 }
 
-} // namespace parser
-} // namespace test
-} // namespace bpftrace
+TEST(Parser, variable_declarations)
+{
+  test("BEGIN { let $x; }", R"(
+Program
+ BEGIN
+  decl
+   variable: $x
+)");
+
+  test("BEGIN { let $x: int8; }", R"(
+Program
+ BEGIN
+  decl
+   variable: $x :: [int8]
+)");
+
+  test("BEGIN { let $x = 1; }", R"(
+Program
+ BEGIN
+  decl
+   variable: $x
+   int: 1
+)");
+
+  test("BEGIN { let $x: int8 = 1; }", R"(
+Program
+ BEGIN
+  decl
+   variable: $x :: [int8]
+   int: 1
+)");
+
+  // Needs the let keyword
+  test_parse_failure("BEGIN { $x: int8; }", R"(
+stdin:1:9-12: ERROR: syntax error, unexpected :, expecting ; or }
+BEGIN { $x: int8; }
+        ~~~
+)");
+
+  // Needs the let keyword
+  test_parse_failure("BEGIN { $x: int8 = 1; }", R"(
+stdin:1:9-12: ERROR: syntax error, unexpected :, expecting ; or }
+BEGIN { $x: int8 = 1; }
+        ~~~
+)");
+}
+
+} // namespace bpftrace::test::parser
