@@ -21,8 +21,7 @@ static long myclock()
 
 int main()
 {
-  while (1)
-  {
+  while (1) {
     myclock();
   }
   return 0;
