@@ -1,19 +1,17 @@
+#include <cstdlib>
+#include <cstring>
+#include <filesystem>
+#include <fstream>
+
 #include "utils.h"
 #include "gmock/gmock.h"
 #include "gtest/gtest.h"
-#include <cstring>
 #include <fcntl.h>
-#include <fstream>
-#include <stdlib.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
 
-#include "filesystem.h"
-
-namespace bpftrace {
-namespace test {
-namespace utils {
+namespace bpftrace::test::utils {
 
 TEST(utils, split_string)
 {
@@ -61,6 +59,40 @@ TEST(utils, split_addrrange_symbol_module)
             tokens_ar_sym_mod);
 }
 
+static void test_erase_parameter_list(std::string input,
+                                      std::string_view expected)
+{
+  erase_parameter_list(input);
+  EXPECT_EQ(input, expected);
+}
+
+TEST(utils, erase_parameter_list)
+{
+  // Trivial cases
+  test_erase_parameter_list("", "");
+  test_erase_parameter_list("()", "");
+  test_erase_parameter_list("void foo", "void foo");
+  test_erase_parameter_list("void foo()", "void foo");
+  test_erase_parameter_list("void foo(Bar &b)", "void foo");
+  // Qualified functions
+  //   we don't need to handle `noexcept` or trailing return type
+  //   because they don't appear in the demangled function name
+  test_erase_parameter_list("void foo() &&", "void foo");
+  test_erase_parameter_list("void foo() const", "void foo");
+  // Templated parameter/function
+  test_erase_parameter_list("void foo(Bar<Baz> &b)", "void foo");
+  test_erase_parameter_list("void foo(Bar<Baz()> &b)", "void foo");
+  test_erase_parameter_list("void foo<Bar()>()", "void foo<Bar()>");
+  test_erase_parameter_list("void foo<Bar()>::foo(Bar &b)",
+                            "void foo<Bar()>::foo");
+  // Function pointer
+  test_erase_parameter_list("void foo(void (*func)(int))", "void foo");
+  test_erase_parameter_list("void foo(void (*func)(int, Bar<Baz()>))",
+                            "void foo");
+  // Missing closing parenthesis
+  test_erase_parameter_list("void foo(Bar &b", "void foo(Bar &b");
+}
+
 TEST(utils, wildcard_match)
 {
   std::vector<std::string> tokens_not = { "not" };
@@ -152,7 +184,7 @@ TEST(utils, resolve_binary_path)
   EXPECT_EQ(resolve_binary_path(path + "/executable*"), paths_all_executables);
   EXPECT_EQ(resolve_binary_path(path + "/*executable*"), paths_all_executables);
 
-  EXPECT_GT(std_filesystem::remove_all(path), 0);
+  EXPECT_GT(std::filesystem::remove_all(path), 0);
 }
 
 TEST(utils, abs_path)
@@ -180,19 +212,23 @@ TEST(utils, abs_path)
   EXPECT_EQ(abs_path(std::string("/proc/1/root/usr/local/bin/usdt_test.so")),
             std::string("/proc/1/root/usr/local/bin/usdt_test.so"));
 
-  EXPECT_TRUE(std_filesystem::remove(rel_file));
-  EXPECT_GT(std_filesystem::remove_all(path), 0);
+  EXPECT_TRUE(std::filesystem::remove(rel_file));
+  EXPECT_GT(std::filesystem::remove_all(path), 0);
 }
 
 TEST(utils, get_cgroup_hierarchy_roots)
 {
   auto roots = get_cgroup_hierarchy_roots();
 
-  // Check that each entry is a proper cgroup filesystem
-  for (auto root : roots) {
-    EXPECT_TRUE(root.first == "cgroup" || root.first == "cgroup2");
-    std_filesystem::path root_path(root.second);
-    EXPECT_TRUE(std_filesystem::exists(root_path / "cgroup.procs"));
+  // Check that each entry is a proper cgroup filesystem. The first set are
+  // cgroupv1 results, and the second set are cgroupv2 results.
+  for (auto root : roots[0]) {
+    std::filesystem::path root_path(root);
+    EXPECT_TRUE(std::filesystem::exists(root_path / "cgroup.procs"));
+  }
+  for (auto root : roots[1]) {
+    std::filesystem::path root_path(root);
+    EXPECT_TRUE(std::filesystem::exists(root_path / "cgroup.procs"));
   }
 }
 
@@ -204,14 +240,14 @@ TEST(utils, get_cgroup_path_in_hierarchy)
     throw std::runtime_error("creating temporary path for tests failed");
   }
 
-  const std_filesystem::path path(tmpdir);
-  const std_filesystem::path file_1 = path / "file1";
-  const std_filesystem::path subdir = path / "subdir";
-  const std_filesystem::path file_2 = subdir / "file2";
+  const std::filesystem::path path(tmpdir);
+  const std::filesystem::path file_1 = path / "file1";
+  const std::filesystem::path subdir = path / "subdir";
+  const std::filesystem::path file_2 = subdir / "file2";
 
   // Make a few files in the directory to imitate cgroup files and get their
   // inodes
-  if (!std_filesystem::create_directory(subdir)) {
+  if (!std::filesystem::create_directory(subdir)) {
     throw std::runtime_error("creating subdirectory for tests failed");
   }
   static_cast<std::ofstream &&>(std::ofstream(file_1) << "File 1 content")
@@ -231,7 +267,7 @@ TEST(utils, get_cgroup_path_in_hierarchy)
               "/subdir/file2");
   }
 
-  EXPECT_GT(std_filesystem::remove_all(tmpdir), 0);
+  EXPECT_GT(std::filesystem::remove_all(tmpdir), 0);
 }
 
 TEST(utils, parse_kconfig)
@@ -294,16 +330,16 @@ TEST(utils, find_in_path)
   ASSERT_TRUE(::mkdtemp(&tmpdir[0]));
 
   // Create some directories
-  const std_filesystem::path path(tmpdir);
-  const std_filesystem::path usr_bin = path / "usr" / "bin";
-  const std_filesystem::path usr_local_bin = path / "usr" / "local" / "bin";
-  ASSERT_TRUE(std_filesystem::create_directories(usr_bin));
-  ASSERT_TRUE(std_filesystem::create_directories(usr_local_bin));
+  const std::filesystem::path path(tmpdir);
+  const std::filesystem::path usr_bin = path / "usr" / "bin";
+  const std::filesystem::path usr_local_bin = path / "usr" / "local" / "bin";
+  ASSERT_TRUE(std::filesystem::create_directories(usr_bin));
+  ASSERT_TRUE(std::filesystem::create_directories(usr_local_bin));
 
   // Create some dummy binaries
-  const std_filesystem::path usr_bin_echo = usr_bin / "echo";
-  const std_filesystem::path usr_local_bin_echo = usr_local_bin / "echo";
-  const std_filesystem::path usr_bin_cat = usr_bin / "cat";
+  const std::filesystem::path usr_bin_echo = usr_bin / "echo";
+  const std::filesystem::path usr_local_bin_echo = usr_local_bin / "echo";
+  const std::filesystem::path usr_bin_cat = usr_bin / "cat";
   {
     std::ofstream(usr_bin_echo) << "zz";
     std::ofstream(usr_local_bin_echo) << "zz";
@@ -344,7 +380,7 @@ TEST(utils, find_in_path)
   });
 
   // Cleanup
-  EXPECT_TRUE(std_filesystem::remove_all(path));
+  EXPECT_TRUE(std::filesystem::remove_all(path));
 }
 
 // These tests are a bit hacky and rely on repository structure.
@@ -356,9 +392,12 @@ TEST(utils, find_in_path)
 TEST(utils, find_near_self)
 {
   auto runtime_tests = find_near_self("runtime-tests.sh");
+  // clang-tidy is not aware ASSERT_*() terminates testcase
+  // NOLINTBEGIN(bugprone-unchecked-optional-access)
   ASSERT_TRUE(runtime_tests.has_value());
   EXPECT_TRUE(runtime_tests->filename() == "runtime-tests.sh");
-  EXPECT_TRUE(std_filesystem::exists(*runtime_tests));
+  EXPECT_TRUE(std::filesystem::exists(*runtime_tests));
+  // NOLINTEND(bugprone-unchecked-optional-access)
 
   EXPECT_FALSE(find_near_self("SHOULD_NOT_EXIST").has_value());
 }
@@ -374,6 +413,18 @@ TEST(utils, get_pids_for_program)
   ASSERT_EQ(pids.size(), 0);
 }
 
-} // namespace utils
-} // namespace test
-} // namespace bpftrace
+TEST(utils, round_up_to_next_power_of_two)
+{
+  // 2^31 = 2147483648 which is max power of 2 within uint32_t
+  constexpr uint32_t max_power_of_two = 2147483648;
+  ASSERT_EQ(round_up_to_next_power_of_two(0), 0);
+  ASSERT_EQ(round_up_to_next_power_of_two(1), 1);
+  ASSERT_EQ(round_up_to_next_power_of_two(7), 8);
+  ASSERT_EQ(round_up_to_next_power_of_two(55), 64);
+  ASSERT_EQ(round_up_to_next_power_of_two(128), 128);
+  ASSERT_EQ(round_up_to_next_power_of_two(max_power_of_two - 1),
+            max_power_of_two);
+  ASSERT_EQ(round_up_to_next_power_of_two(max_power_of_two), max_power_of_two);
+}
+
+} // namespace bpftrace::test::utils
